@@ -22,14 +22,14 @@
 #define TINYPROXY_BASICAUTH_H
 
 #include <stddef.h>
-#include "vector.h"
+#include "sblist.h"
 
 extern ssize_t basicauth_string(const char *user, const char *pass,
 	char *buf, size_t bufsize);
 
-extern void basicauth_add (vector_t authlist,
+extern void basicauth_add (sblist *authlist,
 	const char *user, const char *pass);
 
-extern int basicauth_check (vector_t authlist, const char *authstring);
+extern int basicauth_check (sblist *authlist, const char *authstring);
 
 #endif
